//==============================================================================================
// Module D17 par Jean-Michel Dubois
// Suivi des trains et régulation automatique
//==============================================================================================

#include "config.h"

#if PROTOCOLE == D17PPV3

#include "block.h"
#include "user.h"
#include "inputs.h"
#include "protocol.h"
#include "signal.h"
#include "turnout.h"

Train Block::m_Trains[256];	// Table des trains

#if PRINT_SIGNAL

char* szTypeFeux[] = {
	"ETEINT",
	"VOIE_LIBRE",
	"AVERTISSEMENT",
	"3 ???",
	"SEMAPHORE",
	"CARRE",
	"MANOEUVRE",
	"CARRE_VIOLET"
};

char* szTypeRalenti[]{
	"VITESSE_NORMALE",
	"RALENTISSEMENT_30",
	"RAPPEL_30",
	"RALENTISSEMENT_60",
	"RAPPEL_60"
};

#endif

void Block::InitTco(void)
{
	TRACE_TRACKING("Block::InitTco\n");

	for (byte numsignal = 0; numsignal < m_uiSignaux; ++numsignal)
		ChangeSignal(numsignal, VOIE_LIBRE);

	for (Zone* pZone = m_Zones; pZone < &m_Zones[m_uiZones]; ++pZone)
		if (pZone->EstOccupee())
			OnZoneOccuper(pZone - m_Zones);

	for (byte numaig = 0; numaig < NB_AIGUILLES; ++numaig)
		OnNotifyAig(numaig, AIG_DIRECT);
}

void Block::UpdateZones(void)
{
	// Procède en deux passes, d'abord les zones nouvellement occupées, ensuite les zones nouvellement libérées.
	// Permet de propager d'une zone à l'autre l'adresse dcc des trains/wagons détectés de façon ponctuelle.
	for (byte uiZ = 0; uiZ < m_uiZones; ++uiZ)
		if (user_get_in(uiZ) && !user_get_in_old(uiZ))
			Occuper(uiZ);

	for (byte uiZ = 0; uiZ < m_uiZones; ++uiZ)
		if (!user_get_in(uiZ) && user_get_in_old(uiZ))
			Liberer(uiZ);
}

void Block::Occuper(byte zone)
{
	TRACE_TRACKING("Block::Occuper zone %d\n", zone);
	m_Zones[zone].InsereTrain();	// Recherche le train entrant et met à jour l'état de la zone.
	OnZoneOccuper(zone);			// Modifie la signalisation, fermer un passage à niveau et autres actions.
}

void Block::Liberer(byte zone)
{
	// Un train sortant d'une zone est toujours celui du côté pair car si deux trains sont présents dans une zone,
	// leur départ n'est pas détecté. Or, un train unique est toujours considéré du côté pair.
	Zone* pZone = &m_Zones[zone];
	byte train = pZone->m_cotePair.m_train;
	Train* pTrain = &m_Trains[train];

	TRACE_TRACKING("Block::Liberer %d, train %d\n", zone, train);

	// Zone précédente dans le sens pair
	Zone* pZonePaire = (pZone->m_cotePair.m_zoneSortie >= m_uiZones) ? NULL : &m_Zones[pZone->m_cotePair.m_zoneSortie];
	// Zone précédente dans le sens impair
	Zone* pZoneImpaire = (pZone->m_coteImpair.m_zoneSortie >= m_uiZones) ? NULL : &m_Zones[pZone->m_coteImpair.m_zoneSortie];

	// Si n'y a aucune détection dans les zones voisines, soit il y a une perte de contact, soit le train a été enlevé manuellement.
	// Considérer que le train est toujours présent pour ne pas perde les informations en attendant une confirmation.

	if (pTrain->m_sensActuel == SENS_PAIR) {
		if (!pZonePaire->EstOccupee()) {
			TRACE_TRACKING("Zone::Liberer train %d non retrouvé dans la zone paire %d\n", train, pZone->m_cotePair.m_zoneSortie);
			return;
		}
	} else if (!pZoneImpaire->EstOccupee()) {
		TRACE_TRACKING("Zone::Liberer train %d non retrouvé dans la zone impaire %d\n", train, pZone->m_coteImpair.m_zoneSortie);
		return;
	}

	// Libère les zones à double occupation du train qui vient de quitter la zone.
	// Le passage dans la zone voisine d'un train quittant une zone occupée par deux trains ne déclenche pas
	// d'évènement de libération car la zone quittée détecte toujours la présence d'un train.
	// Donc, lorsqu'un train quitte une zone, on regarde si le même train est présent dans la zone précédente dans le sens de sa marche.
	// S'il y est, il ne peut s'agir que d'un train "oublié".
	if (pTrain->m_sensActuel == SENS_PAIR) {
		if (pZoneImpaire != NULL && pZoneImpaire->m_coteImpair.m_train) {
			if (pZoneImpaire->m_coteImpair.m_train == train) {
				// Supprime le train côté impair
				pZoneImpaire->m_coteImpair.m_train = LIBRE;
				pZoneImpaire->Modifier();
				TRACE_TRACKING("Zone::Liberer %d, train %d enlevé de la zone %d\n", zone, train, pZoneImpaire - m_Zones);
			} else if (pZoneImpaire->m_cotePair.m_train == train) {
				// Passe le train côté impair du côté pair.
				pZoneImpaire->m_cotePair.m_train = pZoneImpaire->m_coteImpair.m_train;
				pZoneImpaire->m_coteImpair.m_train = LIBRE;
				pZoneImpaire->Modifier();
				TRACE_TRACKING("Zone::Liberer %d, train %d enlevé de la zone %d\n", zone, train, pZoneImpaire - m_Zones);
			}
		}
	} else {
		if (pZonePaire != NULL && pZonePaire->m_coteImpair.m_train) {
			// Supprime le train pair
			if (pZonePaire->m_cotePair.m_train == train) {
				// Passe le train côté impair du côté pair.
				pZonePaire->m_cotePair.m_train = pZonePaire->m_coteImpair.m_train;
				pZonePaire->m_coteImpair.m_train = LIBRE;
				pZonePaire->Modifier();
				TRACE_TRACKING("Zone::Liberer %d, train %d enlevé de la zone %d\n", zone, train, pZonePaire - m_Zones);
			}
		}
	}

	// Ajoute le train en mouvement à la zone située dans le sens de sa marche.
	if (pTrain->m_sensActuel == SENS_PAIR) {
		// Si la zone suivante paire est déjà occupée par un train...
		if (pZonePaire->m_cotePair.m_train) {
			// et que le train entrant n'est pas le même...
			if (pZone->m_cotePair.m_train != train) {
				// le faire entrer du côté côté impair de la zone.
				pZonePaire->m_coteImpair.m_train = train;
				// Met à jour la position de la tête du train.
				pTrain->m_zone = pZone->m_cotePair.m_zoneSortie;
				TRACE_TRACKING("Zone::Liberer %d, train %d reporté zone %d côté impair\n", zone, train, pZonePaire - m_Zones);
			}
		} else {
			// Sinon, le train occupe le côté pair de la zone.
			pZonePaire->m_cotePair.m_train = train;
			// Met à jour la position de la tête du train.
			pTrain->m_zone = pZone->m_cotePair.m_zoneSortie;
			TRACE_TRACKING("Zone::Liberer %d, train %d reporté zone %d côté pair\n", zone, train, pZonePaire - m_Zones);
		}
	} else if (pTrain->m_sensActuel == SENS_IMPAIR) {
		// Si la zone suivante impaire est déjà occupée par un autre train...
		if (pZoneImpaire->m_cotePair.m_train && pZoneImpaire->m_cotePair.m_train != train)
			// elle l'est par principe du côté pair. "Pousser" le train déjà présent du côté impair.
			pZoneImpaire->m_coteImpair.m_train = pZoneImpaire->m_cotePair.m_train;

		// Faire rentrer le train en mouvement du côté pair.
		pZoneImpaire->m_cotePair.m_train = train;
		// Met à jour la position de la tête du train.
		pTrain->m_zone = pZone->m_coteImpair.m_zoneSortie;
		TRACE_TRACKING("Zone::Liberer %d, train %d reporté zone %d côté pair\n", zone, train, pZoneImpaire - m_Zones);
	}

	pZone->m_dernier_train = pZone->m_cotePair.m_train;
	pZone->m_cotePair.m_train = LIBRE;
	pZone->Modifier();

	if (CantonOccupe(zone))	// Si une des zones d'un canton est occupée, ne plus rien faire.
		return;

	OnZoneLiberer(zone);	// Modifie la signalisation, ouvre un passage à niveau et autres actions.
}

// Sécurité des circulations. Fait ralentir, stopper et repartir les trains en fonction de l'état des feux et du type de zone.

void Block::Reguler(void)
{
	for (Train* pTrain = &m_Trains[0]; pTrain <= &m_Trains[DERNIER_TRAIN]; ++pTrain) {
		if (!pTrain->m_marche_a_vue && pTrain->m_pasMax && pTrain->m_pasSouris && pTrain->m_zone < m_uiZones) {
			Zone* pZone = &m_Zones[pTrain->m_zone];

			if (pTrain->m_sensActuel == SENS_PAIR) {
				switch (pZone->m_typeZoneSensPair) {
				case ZONE_ARRET:
					pTrain->ArreterOuRalentir(&pZone->m_cotePair);	// Arrêter ou ralentir
					break;
				case ZONE_RALENTI:
					pTrain->Ralentir(&pZone->m_cotePair);	// Ralentir
					break;
				case ZONE_MANOEUVRE:
					if (Block::SignalFerme(pZone->m_cotePair.m_signalSortie))
						pTrain->RegulerVitesse(0);
					else if (pZone->m_cotePair.m_zoneSortie == HEURTOIR)
						pTrain->RegulerVitesse(0);
					else
						pTrain->RegulerVitesse(min(pTrain->m_pasSouris, pTrain->m_pas30));

					break;
				case ZONE_AIGUILLES:
					break;
				}
			} else {
				switch (pZone->m_typeZoneSensImpair) {
				case ZONE_ARRET:
					pTrain->ArreterOuRalentir(&pZone->m_coteImpair);	// Arrêter ou ralentir
					break;
				case ZONE_RALENTI:
					pTrain->Ralentir(&pZone->m_coteImpair);	// Ralentir
					break;
				case ZONE_MANOEUVRE:
					if (Block::SignalFerme(pZone->m_cotePair.m_signalSortie))
						pTrain->RegulerVitesse(0);
					else if (pZone->m_coteImpair.m_zoneSortie == HEURTOIR)
						pTrain->RegulerVitesse(0);
					else
						pTrain->RegulerVitesse(min(pTrain->m_pasSouris, pTrain->m_pas30));

					break;
				case ZONE_AIGUILLES:
					break;
				}
			}
		}
	}
}

bool Block::SignalFerme(byte nosignal)
{
	byte f = EtatSignal(nosignal) & ~CLIGNOTANT;

	return (f == SEMAPHORE || f == CARRE || f == CARRE_VIOLET);
}

bool Block::SignalOuvert(byte nosignal)
{
	byte f = EtatSignal(nosignal) & ~CLIGNOTANT;

	return (f == VOIE_LIBRE || f == MANOEUVRE || f == AVERTISSEMENT);
}

bool Block::SignalAvertissement(byte nosignal)
{
	return ((EtatSignal(nosignal) & ~CLIGNOTANT) == AVERTISSEMENT);
}

bool Block::Signal15kmh(byte nosignal)
{
	return EtatSignal(nosignal) == SEMAPHORE_CLIGNO;
}

bool Block::Signal160kmh(byte nosignal)
{
	return EtatSignal(nosignal) == VOIE_LIBRE_CLIGNO;
}

void Block::ChangeSignal(unsigned int numsignal, typeFeux feux, typeRalenti ralenti)
{
#if PRINT_SIGNAL
	extern char* szNomSignal[];

	TRACE_TRACKING("ChangeSignal %s : m_Feux %s%s, m_Ralenti %s\n",
		numsignal < m_uiSignaux ? szNomSignal[numsignal] : "???",
		(feux & 0x07) < 8 ? szTypeFeux[feux & 0x07] : "???",
		(feux & 0x10) ? " CLIGNOTANT" : "",
		ralenti < 5 ? szTypeRalenti[ralenti] : "???");
#endif
	if (numsignal >= m_uiSignaux)
		return;

	Signal* pSignal = &m_Signaux[numsignal];

	if (pSignal->m_Feux != feux || pSignal->m_Ralenti != ralenti) {
		pSignal->m_Feux = feux;
		pSignal->m_Ralenti = ralenti;
		(*pSignal->m_pSignal)(pSignal);

		for (byte i = 0; i < MAX_WIFI_CLIENTS; ++i)
			pSignal->m_fModifie[i] = true;
	}
}

byte Block::EtatSignal(unsigned int numsignal)
{
	return (numsignal >= m_uiSignaux) ? ETEINT : m_Signaux[numsignal].m_Feux;
}

byte Block::EtatRalentissement(unsigned int numsignal)
{
	return (numsignal >= m_uiSignaux) ? VITESSE_NORMALE : m_Signaux[numsignal].m_Ralenti;
}

Train* Block::FindTrainByAddress(unsigned int adresse)
{
	if (adresse < 127)
		return &m_Trains[adresse];

	byte i;
	Train* pTrain;

	for (pTrain = m_Trains; pTrain < &m_Trains[255]; ++pTrain)
		if (pTrain->m_adresse == adresse)
			return pTrain;

	return NULL;
}

#endif
