//----------------------------------------------------------------------------------
// Module D17 par Jean-Michel Dubois
// MCP23017 input and output via bus I2C
//----------------------------------------------------------------------------------

#include "config.h"

#if USER_USE_MCPIN || USER_USE_MCPOUT
# include "mcp23017.h"
# include "i2c.h"

constexpr auto MCP23017_IODIRA = 0x00;
constexpr auto MCP23017_GPINTENA = 0x04;
constexpr auto MCP23017_DEFVALA = 0x06;
constexpr auto MCP23017_GPPUA = 0x0C;
constexpr auto MCP23017_INTFA = 0x0E;
constexpr auto MCP23017_INTFB = 0x0F;
constexpr auto MCP23017_GPIOA = 0x12;
constexpr auto MCP23017_GPIOB = 0x13;
constexpr auto MCP23017_BASE = 0x40;

static bool i2c_mcp23017_online[USER_MCP23017_IN_MODULES + USER_MCP23017_OUT_MODULES];

static bool mcp_wr_16(byte i2c_adr, byte reg, unsigned short val)
{
	// select address
	i2c_start();
	if (!i2c_wr_8(MCP23017_BASE + 2 * i2c_adr + I2C_WRITE))   // 0100AAAW
	{
		TRACE1("mcp_wr_16 1 %d noack\n", i2c_adr);
		return false;
	}
	// select registers
	if (!i2c_wr_8(reg))
	{
		TRACE1("mcp_wr_16 2 %d noack\n", i2c_adr);
		return false;
	}
	// write register A
	if (!i2c_wr_8(val & 0xff))
	{
		TRACE1("mcp_wr_16 3 %d noack\n", i2c_adr);
		return false;
	}
	// write register B
	if (!i2c_wr_8(val >> 8))
	{
		TRACE1("mcp_wr_16 4 %d noack\n", i2c_adr);
		return false;
	}
	i2c_stop();
	return true;
}

#endif

#if USER_USE_MCPIN

# include "inputs.h"
# include "dcc.h"

static int mcp_rd_8(byte i2c_adr, byte reg)
{
	byte v;

	// select address
	i2c_start();
	if (!i2c_wr_8(MCP23017_BASE + 2 * i2c_adr + I2C_WRITE))   // 0100AAAW
	{
		TRACE1("mcp_rd_8 1 %d noack\n", i2c_adr);
		return -1;
	}
	// select register
	if (!i2c_wr_8(reg))
	{
		TRACE1("mcp_rd_8 2 %d noack\n", i2c_adr);
		return -1;
	}
	i2c_stop();
	// read port A
	i2c_start();
	if (!i2c_wr_8(MCP23017_BASE + 2 * i2c_adr + I2C_READ))   // 0100AAAR
	{
		TRACE1("mcp_rd_8 3 %d noack\n", i2c_adr);
		return -1;
	}

	v = i2c_rd_8();
	i2c_wr_ack(0);
	i2c_stop();
	return (int)v;
}

// Look for available mcp23017 modules

void in_mcp23017_init(void)
{
	byte i2c_adr;

	for (i2c_adr = 0; i2c_adr < USER_MCP23017_IN_MODULES; ++i2c_adr)
	{
		// configure port direction
		if (!mcp_wr_16(i2c_adr, MCP23017_IODIRA, 0xffff))
			continue;

		i2c_mcp23017_online[i2c_adr] = true;
		TRACE1(" in mcp23017 %d found\n", i2c_adr);
		// activate pull up 
		if (!mcp_wr_16(i2c_adr, MCP23017_GPPUA, 0xffff))
			continue;

		yield();
	}
}

void in_mcp23017_maj(void)
{
	int v, w;
	byte i2c_adr, i, j, client;

	for (i2c_adr = 0, i = USER_S88_MODULES * 2, j = i + 1; i2c_adr < USER_MCP23017_IN_MODULES; ++i2c_adr, i += 2, j += 2)
	{
		if (!i2c_mcp23017_online[i2c_adr])
			continue;

		io_in_old[i] = io_in[i];
		io_in_old[j] = io_in[j];

		if ((v = mcp_rd_8(i2c_adr, MCP23017_GPIOA)) == -1)
			continue;

		v ^= 0xff;
		io_in[i] = (byte)v;

		for (client = 0; client < MAX_WIFI_CLIENTS; ++client)
			io_inmem_wifi[client][i] |= (byte)v;

		if ((w = mcp_rd_8(i2c_adr, MCP23017_GPIOB)) == -1)
			continue;

		w ^= 0xff;
		io_in[j] = (byte)w;

		for (client = 0; client < MAX_WIFI_CLIENTS; ++client)
			io_inmem_wifi[client][j] |= (byte)w;

		yield();
		//    TRACE1("in_mcp23017_maj read %d ok, value %02x%02x\n", i, v, w);
	}
}

#endif

#if USER_USE_MCPOUT

static bool mcp_wr_8(byte i2c_adr, byte reg, byte val)
{
	// select address
	i2c_start();
	if (!i2c_wr_8(MCP23017_BASE + 2 * i2c_adr + I2C_WRITE))   // 0100AAAW
	{
		TRACE1("mcp_wr_8 1 %d noack\n", i2c_adr);
		return false;
	}
	// select registers
	if (!i2c_wr_8(reg))
	{
		TRACE1("mcp_wr_8 2 %d noack\n", i2c_adr);
		return false;
	}
	// write register
	if (!i2c_wr_8(val))
	{
		TRACE1("mcp_wr_8 3 %d noack\n", i2c_adr);
		return false;
	}
	i2c_stop();
	return true;
}

// Look for available mcp23017 modules

static byte out_mcp23017_pins[2 * USER_MCP23017_OUT_MODULES];

void out_mcp23017_init(void)
{
	byte i2c_adr;

	for (i2c_adr = USER_MCP23017_IN_MODULES; i2c_adr < USER_MCP23017_IN_MODULES + USER_MCP23017_OUT_MODULES; ++i2c_adr)
	{
		// configure port direction
		if (!mcp_wr_16(i2c_adr, MCP23017_IODIRA, 0))
			continue;

		i2c_mcp23017_online[i2c_adr] = true;
		TRACE1(" out mcp23017 %d found\n", i2c_adr);
		yield();
	}

	memset(out_mcp23017_pins, 0, sizeof(out_mcp23017_pins));
}

// change MCP23017 output
// num = pin number from 16 to 143 -> 0 to 127
// val = 0, 1

void out_mcp23017(byte num, byte val)
{
	if ((num -= MCP23017_OUT_PIN_OFFSET) >= 16 * USER_MCP23017_OUT_MODULES)
		return;

	byte i2c_adr = num / 16;

	if (!i2c_mcp23017_online[i2c_adr])
		return;

	byte* index = &out_mcp23017_pins[num / 8];

	if (val)
		*index |= (1 << (num % 8));
	else
		*index &= ~(1 << (num % 8));

	mcp_wr_8(i2c_adr, MCP23017_GPIOA + (num % 16) / 8, *index);
	TRACE1("out_mcp23017 i2c adr %d, register %d, value %d\n", i2c_adr, (num % 16) / 8, *index);
	yield();
}
#endif