//==============================================================================================
// Module D17 par Jean-Michel Dubois
// Schéma logique du réseau correspondant aux TCOs
//==============================================================================================

#include "config.h"

#if PROTOCOLE == D17PPV3 

#include "block.h"
#include "user.h"
#include "signal.h"
#include "Zone.h"
#include "Train.h"
#include "d19.h"
#include "commands.h"
#include "turnout.h"

// La configuration réseau IP de la centrale, les adresses et types des décodeurs d'accessoires, 
// les détecteurs d'occupation S88 ou d'état doivent être décrits dans config.h.


// La configuration du réseau ferroviaire est décrite dans ce fichier.
// Toutes les tables et fonctions correspondent au réseau décrit dans le fichier tco1.txt fourni avec le logiciel SourisD17.

// ---------- DESCRIPTION DES SIGNAUX LOGIQUES

// Noms et numéros des signaux logiques. Doit correspondre aux signaux nécessaires à la circulation en sécurité. 
// Les noms utilisés sont libres. La numérotation est automatique partant de 0 jusqu'à 254. 
// Les signaux peuvent ne pas exister physiquement.

enum {
	SIGNAL_PAIR_0_1,	// sémaphore
	SIGNAL_PAIR_2_3,	// sémaphore
	SIGNAL_PAIR_4_5,	// carré
	SIGNAL_IMPAIR_1_0,	// carré + rappel 30km/h
	SIGNAL_IMPAIR_5_4,	// sémaphore
	SIGNAL_IMPAIR_3_2,	// sémaphore + ralentissement 30
	SIGNAL_PAIR_7,		// carré violet
	NB_SIGNAUX,			// doit toujours être l'avant-dernière ligne de la liste.
	AUCUN = 255			// doit toujours être la dernière ligne de la liste. Utilisé pour indiquer l'absence de signal en sortie de zone.
};

#if PRINT_SIGNAL
// Table des noms de signaux pour le débogage

char* szNomSignal[] = {
	"SIGNAL_PAIR_0_1",		// sémaphore
	"SIGNAL_PAIR_2_3",		// sémaphore
	"SIGNAL_PAIR_4_5",		// carré
	"SIGNAL_IMPAIR_1_0",	// carré + rappel 30km/h
	"SIGNAL_IMPAIR_5_4",	// sémaphore
	"SIGNAL_IMPAIR_3_2",	// sémaphore + ralentissement 30
	"SIGNAL_PAIR_7"			// carré violet
};
#endif

// ---------- DESCRIPTION DES SIGNAUX PHYSIQUES

// Commandes de la signalisation physique.
// Voir signal.h et d19.h pour la signification des valeurs.
// D'autres modules de gestion de signaux peuvent être créés.
// L'ordre des informations doit être le même que celui des signaux logiques.
//
// Paramètres de chaque signal :
//  - nom de la fonction de commande des signaux physiques.
//  - signal connecté au bus DCC : adresse DCC étendue. D19 : numéro de module MAX7219 sur bus SPI.
//  - signal connecté au bus DCC : inutilisé. D19 : numéro du signal du module.

Signal Block::m_Signaux[] = {
	{ d19::a3feux, 0, 0 }, // SIGNAL_PAIR_0_1
	{ d19::a3feux, 0, 1 }, // SIGNAL_PAIR_2_3
	{ d19::a5feux, 0, 0 }, // SIGNAL_PAIR_4_5
	{ d19::a5feux, 0, 1 }, // SIGNAL_IMPAIR_1_0
	{ d19::a3feux, 0, 2 }, // SIGNAL_IMPAIR_5_4
	{ d19::a5feux, 0, 2 }, // SIGNAL_IMPAIR_3_2
	{ d19::a3feux, 0, 3 }  // SIGNAL_PAIR_7
};

byte Block::m_uiSignaux = NB_SIGNAUX;	// Ne pas modifier, ne pas supprimer.

// ---------- DESCRIPTION DES ZONES

// Zones de détection. Doit correspondre aux zones réellement présentes.
// Toutes les zones doivent être munies de détecteurs d'occupation.
// Leur numérotation va de 0 à 254. Elle est automatiquement créée par la position dans la table.
// La première ligne décrit la zone 0, la seconde la zone 1 et ainsi de suite.
// Le numéro 255 est réservé aux heurtoirs.
// Le numéro des zones de sortie des zones d'arrêt, de ralentissement et de manoeuvre est immuable.
// Ceux des zones d'aiguilles est modifié dans la fonction Block::OnNotifyAig en fonction de la direction de l'aiguille.

// Paramètres de chaque zone :
// - type dans le sens pair.
//   ZONE_RALENTI	: zone de ralentissement
//   ZONE_ARRET		: zone d'arrêt
//   ZONE_AIGUILLES	: zone d'aiguilles
//   ZONE_MANOEUVRE	: zone de manoeuvre
//   Le type de zone est utilisé par la régulation automatique de la vitesse des trains.
// - numéro de la prochaine zone dans le sens pair. Variable quand la zone décrite est une zone d'aiguilles, fixe pour les autres types. 255 = heurtoir.
// - nom du signal contrôlant l'accès à la prochaine zone dans le sens pair. 
//   La valeur AUCUN = pas de signal  est normalement réservée aux zones d'aiguilles et aux zones se terminant par un heurtoir.
// - vitesse maximale dans la zone : V15, V30, V60, V90, V100, V120, V160, V220, V300 et V320.
// - type dans le sens impair.
// - numéro de la prochaine zone dans le sens pair. 
// - nom du signal contrôlant l'accès à la prochaine zone dans le sens impair. 

Zone Block::m_Zones[] = {
//	  sens pair							         sens impair
//    type zone, sortie, signal,		  vmax,  type zone,        sortie, signal             vmax
	{ ZONE_RALENTI,   1, SIGNAL_PAIR_0_1, VM160, ZONE_ARRET,            6, SIGNAL_IMPAIR_1_0, VM160 }, // zone 0
	{ ZONE_ARRET,     2, SIGNAL_PAIR_0_1, VM160, ZONE_RALENTI,          0, SIGNAL_IMPAIR_1_0, VM160 }, // zone 1
	{ ZONE_RALENTI,   3, SIGNAL_PAIR_2_3, VM160, ZONE_ARRET,            1, SIGNAL_IMPAIR_3_2, VM160 }, // zone 2
	{ ZONE_ARRET,     4, SIGNAL_PAIR_2_3, VM160, ZONE_RALENTI,          2, SIGNAL_IMPAIR_3_2, VM160 }, // zone 3
	{ ZONE_RALENTI,   5, SIGNAL_PAIR_4_5, VM160, ZONE_ARRET,            3, SIGNAL_IMPAIR_5_4, VM160 }, // zone 4
	{ ZONE_ARRET,     6, SIGNAL_PAIR_4_5, VM160, ZONE_RALENTI,          4, SIGNAL_IMPAIR_5_4, VM160 }, // zone 5
	{ ZONE_AIGUILLES, 0, AUCUN,           VM160, ZONE_AIGUILLES,        5, AUCUN,             VM160 }, // zone 6, initialiser les zones paire/impaire en mode direct
	{ ZONE_MANOEUVRE, 6, SIGNAL_PAIR_7,   VM160, ZONE_MANOEUVRE, HEURTOIR, AUCUN,             VM160 }  // zone 7
};

byte Block::m_uiZones = sizeof(m_Zones) / sizeof(Zone);	// Ne pas modifier, ne pas supprimer.

// ---------- GESTION GLOBALE DES RESSOURCES UTILISEES PAR LES TCO

// Opérations à effectuer lors de l'initialisation de la centrale.
// Exécuté après toutes les initialisations sauf init_wifi et OnInitTco.

void Block::OnInit(void)
{
	// ne pas enlever
	/*
SERVO_INIT(0, 0, 40);
*/
}

// Initialisation de la gestion des TCO. Exécuté après la connexion du premier logiciel client.

void Block::OnInitTco(void)
{
	;
}

// ---------- GESTION DES AIGUILLAGES

// Actionne un aiguillage si les conditions requises sont remplies et gère les conséquences
// de son changement de direction sur l'état des feux et le numéro des zones de sortie.
// Paramètres : numéro d'aiguillage, commande 0=direct 1=dévié

void Block::OnNotifyAig(byte naig, byte cmd)
{
	// Teste les conditions requises pour autoriser le changement.

	// Si un train est présent dans la zone d'aiguilles ou qu'un intinéraire passant par la zone d'aiguille 
	// est constitué, les aiguillages de l'itinéraire ne peuvent pas être basculés.

	if (naig == 0 && ! ZoneOccupee(6)) {
		if (user_get_u(0) || user_get_u(1) || user_get_u(2) || user_get_u(3))
			return;
	}

	// Change la direction de l'aiguillage et le numéro des zones de sortie.

	BasculeAiguille(naig, cmd);

	// Gère les conséquences sur l'état des feux et modifie les zones de sorties des zones d'aiguilles.
	// A écrire pour chaque aiguillage ou groupe d'aiguillages manoeuvrés simultanément.

	if (naig == 0) {
		if (cmd == 0) {	// aiguille directe
			ChangeSignal(SIGNAL_PAIR_7, CARRE_VIOLET);

			if (CantonOccupe(0)) {
				ChangeSignal(SIGNAL_PAIR_4_5, SEMAPHORE);

				if (SignalOuvert(SIGNAL_PAIR_2_3))
					ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);
			} else
				ChangeSignal(SIGNAL_PAIR_4_5, SignalFerme(SIGNAL_PAIR_0_1) ? AVERTISSEMENT : VOIE_LIBRE);

			if (CantonOccupe(4)) {
				ChangeSignal(SIGNAL_IMPAIR_1_0, SEMAPHORE);
				ChangeSignal(SIGNAL_IMPAIR_3_2, CantonOccupe(0) ? SEMAPHORE : AVERTISSEMENT);
			} else {
				ChangeSignal(SIGNAL_IMPAIR_1_0, CantonOccupe(2) ? AVERTISSEMENT : VOIE_LIBRE);

				if (CantonOccupe(0)) {
					ChangeSignal(SIGNAL_IMPAIR_3_2, SEMAPHORE);
					ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);
				} else {
					ChangeSignal(SIGNAL_IMPAIR_3_2, VOIE_LIBRE);
					ChangeSignal(SIGNAL_PAIR_2_3, VOIE_LIBRE);
				}
			}
		} else {
			ChangeSignal(SIGNAL_PAIR_4_5, CARRE);

			if (SignalOuvert(SIGNAL_PAIR_2_3))
				ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);

			if (ZoneOccupee(7)) {
				ChangeSignal(SIGNAL_IMPAIR_1_0, SEMAPHORE_CLIGNO, RAPPEL_30);
				ChangeSignal(SIGNAL_IMPAIR_3_2, CantonOccupe(0) ? SEMAPHORE : AVERTISSEMENT, RALENTISSEMENT_30);
			} else {
				ChangeSignal(SIGNAL_IMPAIR_1_0, MANOEUVRE_CLIGNO, RAPPEL_30);
				ChangeSignal(SIGNAL_IMPAIR_3_2, CantonOccupe(0) ? SEMAPHORE : VOIE_LIBRE, RALENTISSEMENT_30);
			}

			ChangeSignal(SIGNAL_PAIR_7, CantonOccupe(0) ? CARRE_VIOLET : MANOEUVRE);
		}
	}
}

// Actionne physiquement un aiguillage et change les zones de sortie. Voir la documentation de D17.
// L'utilisation des macros de commands.h est recommandée.
// Ne vérifie pas la position réelle des aiguilles.
// Paramètres : numéro d'aiguillage, commande 0=direct 1=dévié

void Block::BasculeAiguille(byte naig, byte cmd)
{
	// Adresses des décodeurs d'accessoires. Doit correspondre aux décodeurs réellement présents.
	// La liste suivante est donnée à titre d'exemple.

	constexpr auto DECODER_ULN2803A_1 = 10; // décodeur intégré à la centrale pour moteurs à bobines
	constexpr auto DECODER_ULN2803A_2 = 11;
	constexpr auto DECODER_L293D_1 = 12;	// décodeur des ponts en H L293D pour moteurs Conrad
	constexpr auto DECODER_L293D_2 = 13;
	constexpr auto PCA9685_A0 = 0;			// premier module PCA9685 sur bus I2C pour servomoteurs

	TRACE_TRACKING("Block::BasculeAiguille %d, cmd %d\n", naig, cmd);
	// Exemples de commandes :
	//  DIRECT_SINGLE_PULSE(acc, pin1, cmd, timer, duration);
	//	SERVO_SINGLE(2, PCA9685_A0, cmd);
	//	DCC_BASIC(3, DECODER_L293D_1, 0, cmd);
	// Obligatoire : Signale le changement de direction aux souris clientes.
	user_set_d_e(naig, cmd ^ 1, cmd);

	if (naig == 0) {
		// Commande l'aiguillage.
		DCC_BASIC(0, DECODER_ULN2803A_1, 0, cmd);
		// Obligatoire : Indique la zone de sortie en fonction de la direction de l'aiguillage.
		if (cmd == AIG_DIRECT)
			m_Zones[6].m_coteImpair.m_zoneSortie = 5;
		else
			m_Zones[6].m_coteImpair.m_zoneSortie = 7;
	}
}

// Fin de temporisation pour les aiguillages temporisés raccordés directement à la centrale D17.
// ATTENTION : un oubli peut griller un aiguillage démuni d'interrupteurs de fin de course.
// Paramètres : numéro de la temporisation.

void Block::OnNotifyTempoEnd(byte num_tempo)
{
	// DIRECT_SINGLE_PULSE_END(timer, pin1);
}

// ---------- GESTION DES BOUTONS DE COMMANDE

// Gère le changement d'état d'un bouton du TCO.
// Peut être utilisé pour commander des accessoires tels que des dételeurs, de l'éclairage ou tout autre accessoire.
// Est également utilisé pour appeler la gestion des itinéraires.
// Paramètres : numéro du bouton, valeur 0, relâché ou 1, enfoncé.

void Block::OnNotifyU(byte num, byte val)
{
	TRACE_TRACKING("Block::OnNotifyU %d %s\n", num, val ? "former " : "deformer");

	if (val)
		TracerItineraire(num);
	else
		DetruireItineraire(num);
}

// ---------- GESTION DES ITINERAIRES

// Liste des itinéraires.

enum {
	Z7_Z0,
	Z0_Z7,
	Z5_Z0,
	Z0_Z5,
};

// Réservation d'itinéraire.
// Bascule les aiguillages et positionne les signaux protégeant un itinéraire. 
// Leur utilisation est facultative mais la fonction doit exister.
// Paramètre : numéro d'itinéraire.

void Block::TracerItineraire(byte itineraire)
{
	// Avant tout, considérer que la formation ne peut avoir lieu.
	user_set_u(itineraire, 0);

	switch (itineraire) {
	case Z7_Z0: // z7 -> z0
		// Si un train est présent dans la zone d'aiguilles, ignore la demande.
		if (ZoneOccupee(6))
			return;

		// Si un itinéraire déjà réservé emprunte les mêmes zones d'aiguilles, ignore la demande.
		if (user_get_u(Z0_Z7) || user_get_u(Z5_Z0) || user_get_u(Z0_Z5))
			return;

		// bascule les aiguilles...
		BasculeAiguille(0, AIG_DEVIE);
		// modifie la signalisation...
		ChangeSignal(SIGNAL_PAIR_7, CantonOccupe(0) ? CARRE_VIOLET : MANOEUVRE);
		ChangeSignal(SIGNAL_PAIR_4_5, CARRE);

		if (SignalOuvert(SIGNAL_PAIR_2_3))
			ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);

		ChangeSignal(SIGNAL_IMPAIR_1_0, CARRE);
		ChangeSignal(SIGNAL_IMPAIR_3_2, SEMAPHORE);

		if (SignalOuvert(SIGNAL_IMPAIR_5_4))
			ChangeSignal(SIGNAL_IMPAIR_5_4, AVERTISSEMENT);

		break;
	case Z0_Z7: // z0 -> z7
		if (ZoneOccupee(6))
			return;

		if (user_get_u(Z7_Z0) || user_get_u(Z5_Z0) || user_get_u(Z0_Z5))
			return;

		BasculeAiguille(0, AIG_DEVIE);
		ChangeSignal(SIGNAL_PAIR_7, CARRE_VIOLET);
		ChangeSignal(SIGNAL_PAIR_4_5, CARRE);

		if (SignalOuvert(SIGNAL_PAIR_2_3))
			ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);

		if (ZoneOccupee(7)) {
			ChangeSignal(SIGNAL_IMPAIR_1_0, SEMAPHORE_CLIGNO, RAPPEL_30);
			ChangeSignal(SIGNAL_IMPAIR_3_2, CantonOccupe(0) ? SEMAPHORE : AVERTISSEMENT, RALENTISSEMENT_30);
		} else {
			ChangeSignal(SIGNAL_IMPAIR_1_0, MANOEUVRE_CLIGNO, RAPPEL_30);
			ChangeSignal(SIGNAL_IMPAIR_3_2, CantonOccupe(0) ? SEMAPHORE : VOIE_LIBRE, RALENTISSEMENT_30);
		}

		ChangeSignal(SIGNAL_PAIR_7, CARRE_VIOLET);
		break;
	case Z5_Z0: // z5 -> z0
		if (ZoneOccupee(6))
			return;

		if (user_get_u(Z7_Z0) || user_get_u(Z0_Z7) || user_get_u(Z0_Z5))
			return;

		BasculeAiguille(0, AIG_DIRECT);
		ChangeSignal(SIGNAL_PAIR_7, CARRE_VIOLET);

		if (CantonOccupe(0)) {
			ChangeSignal(SIGNAL_PAIR_2_3, CantonOccupe(4) ? SEMAPHORE : AVERTISSEMENT);
			ChangeSignal(SIGNAL_PAIR_4_5, SEMAPHORE);
			ChangeSignal(SIGNAL_IMPAIR_3_2, SEMAPHORE);

			if (SignalOuvert(SIGNAL_IMPAIR_5_4))
				ChangeSignal(SIGNAL_IMPAIR_5_4, AVERTISSEMENT);
		} else
			ChangeSignal(SIGNAL_PAIR_2_3, CantonOccupe(4) ? SEMAPHORE : VOIE_LIBRE);

		ChangeSignal(SIGNAL_PAIR_4_5, CantonOccupe(2) ? AVERTISSEMENT : VOIE_LIBRE);
		ChangeSignal(SIGNAL_IMPAIR_3_2, SEMAPHORE);

		if (SignalOuvert(SIGNAL_IMPAIR_5_4))
			ChangeSignal(SIGNAL_IMPAIR_5_4, AVERTISSEMENT);

		break;
	case Z0_Z5: // z0 -> z5
		if (ZoneOccupee(6))
			return;

		if (user_get_u(Z7_Z0) || user_get_u(Z0_Z7) || user_get_u(Z5_Z0))
			return;

		BasculeAiguille(0, AIG_DIRECT);
		ChangeSignal(SIGNAL_PAIR_7, CARRE_VIOLET);

		if (CantonOccupe(4)) {
			ChangeSignal(SIGNAL_IMPAIR_1_0, SEMAPHORE);

			if (SignalOuvert(SIGNAL_IMPAIR_3_2))
				ChangeSignal(SIGNAL_IMPAIR_3_2, AVERTISSEMENT);
		} else if (SignalFerme(SIGNAL_IMPAIR_5_4)) {
			ChangeSignal(SIGNAL_IMPAIR_1_0, AVERTISSEMENT);
		} else
			ChangeSignal(SIGNAL_IMPAIR_1_0, VOIE_LIBRE);


		ChangeSignal(SIGNAL_PAIR_2_3, SEMAPHORE);

		if (SignalOuvert(SIGNAL_PAIR_0_1))
			ChangeSignal(SIGNAL_PAIR_0_1, AVERTISSEMENT);

		break;
	default:
		return;
	}
	// Informer les souris de la réservation.
	user_set_u(itineraire, 1);
}

// Libération des itinéraires.
// Leur utilisation est facultative mais la fonction doit exister.
// Paramètre : numéro d'itinéraire.

void Block::DetruireItineraire(byte itineraire)
{
	// Avant tout, considérer que la libération ne peut avoir lieu.
	user_set_u(itineraire, 1);

	switch (itineraire) {
	case Z7_Z0: // z7 -> z0
	case Z0_Z7: // z0 -> z7
	case Z5_Z0: // z5 -> z0
	case Z0_Z5: // z0 -> z5
		// Si un train est présent dans la zone d'aiguilles, ignore la demande.
		if (ZoneOccupee(6))
			return;

		// bascule les aiguilles...
		BasculeAiguille(0, AIG_DIRECT);
		// modifie la signalisation...
		ChangeSignal(SIGNAL_PAIR_7, CARRE_VIOLET);

		if (CantonOccupe(0)) {
			if (EtatSignal(SIGNAL_PAIR_4_5) == CARRE)
				ChangeSignal(SIGNAL_PAIR_4_5, SEMAPHORE);
		} else {
			ChangeSignal(SIGNAL_IMPAIR_3_2, CantonOccupe(4) ? AVERTISSEMENT : VOIE_LIBRE);
			ChangeSignal(SIGNAL_PAIR_4_5, CantonOccupe(2) ? AVERTISSEMENT : VOIE_LIBRE);
		}
			
		if (CantonOccupe(4)) {
			if (EtatSignal(SIGNAL_IMPAIR_1_0) == CARRE)
				ChangeSignal(SIGNAL_IMPAIR_1_0, SEMAPHORE);
		} else {
			ChangeSignal(SIGNAL_IMPAIR_1_0, CantonOccupe(2) ? AVERTISSEMENT : VOIE_LIBRE);
			ChangeSignal(SIGNAL_PAIR_2_3, CantonOccupe(0) ? AVERTISSEMENT : VOIE_LIBRE);
		}

		break;
	default:
		return;
	}
	// Informer les souris de la libération.
	user_set_u(itineraire, 0);
}

// ---------- GESTION DE L'OCCUPATION DES ZONES ET CANTONS

// Actions exécutées lorsqu'une zone passe à l'état occupé.
// Doit inclure au minimum le changement d'état des signaux.
// En général, l'occupation d'une zone a des répercussions sur l'état des signaux des deux zones aval.
// Le bon endroit pour fermer un passage à niveau.
// Paramètre : numéro de zone venant d'être occupée.

void Block::OnZoneOccuper(byte zone)
{
	TRACE_TRACKING("Block::OnZoneOccuper %d\n", zone);

	switch (zone) {
	case 0:
	case 1:
	case 6:
		// protection sens pair
		if (SignalOuvert(SIGNAL_PAIR_4_5))
			ChangeSignal(SIGNAL_PAIR_4_5, SEMAPHORE);

		if (SignalOuvert(SIGNAL_PAIR_2_3))
			ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);
		// protection sens impair
		ChangeSignal(SIGNAL_IMPAIR_3_2, SEMAPHORE, user_get_aig(0) == AIG_DEVIE ? RALENTISSEMENT_30 : VITESSE_NORMALE);

		if (SignalOuvert(SIGNAL_IMPAIR_5_4))
			ChangeSignal(SIGNAL_IMPAIR_5_4, AVERTISSEMENT);

		break;
	case 2:
	case 3:
		// protection sens pair
		ChangeSignal(SIGNAL_PAIR_0_1, SEMAPHORE);

		if (SignalOuvert(SIGNAL_PAIR_4_5))
			ChangeSignal(SIGNAL_PAIR_4_5, AVERTISSEMENT);
		// protection sens impair
		if (SignalOuvert(SIGNAL_IMPAIR_5_4))
			ChangeSignal(SIGNAL_IMPAIR_5_4, SEMAPHORE);

		if (user_get_aig(0) == AIG_DIRECT && SignalOuvert(SIGNAL_IMPAIR_1_0))
			ChangeSignal(SIGNAL_IMPAIR_1_0, AVERTISSEMENT);

		break;
	case 4:
	case 5:
		// protection sens pair
		if (SignalOuvert(SIGNAL_PAIR_2_3))
			ChangeSignal(SIGNAL_PAIR_2_3, SEMAPHORE);

		if (SignalOuvert(SIGNAL_PAIR_0_1))
			ChangeSignal(SIGNAL_PAIR_0_1, AVERTISSEMENT);
		// protection sens impair
		if (SignalOuvert(SIGNAL_IMPAIR_1_0))
			ChangeSignal(SIGNAL_IMPAIR_1_0, user_get_aig(0) == AIG_DEVIE ? CARRE : SEMAPHORE);

		if (SignalOuvert(SIGNAL_IMPAIR_3_2))
			ChangeSignal(SIGNAL_IMPAIR_3_2, AVERTISSEMENT);

		break;
	case 7:
		if (user_get_aig(0) == AIG_DEVIE)
			ChangeSignal(SIGNAL_IMPAIR_1_0, CARRE, RAPPEL_30);

		break;
	}
}

// Actions exécutées lorsqu'une zone passe à l'état libre.
// Doit inclure au minimum le changement d'état des signaux.
// En général, la libération d'une zone a des répercussions sur l'état des signaux des deux zones amont.
// Le bon endroit pour ouvrir un passage à niveau.
// Paramètre : numéro de zone venant d'être libérée.

#define nobreak

void Block::OnZoneLiberer(byte zone)
{
	TRACE_TRACKING("Block::OnZoneLiberer %d\n", zone);

	switch (zone) {
	case 0:
	case 1:
	case 6:
		if (user_get_aig(0) == AIG_DIRECT) {
			if (CantonOccupe(2)) {
				ChangeSignal(SIGNAL_PAIR_4_5, AVERTISSEMENT);
			} else {
				ChangeSignal(SIGNAL_PAIR_4_5, VOIE_LIBRE);
				ChangeSignal(SIGNAL_IMPAIR_5_4, VOIE_LIBRE);
			}

			if (CantonOccupe(4)) {
				ChangeSignal(SIGNAL_IMPAIR_3_2, AVERTISSEMENT);
				ChangeSignal(SIGNAL_PAIR_2_3, SEMAPHORE);
			} else {
				ChangeSignal(SIGNAL_IMPAIR_3_2, VOIE_LIBRE);
				ChangeSignal(SIGNAL_PAIR_2_3, VOIE_LIBRE);
			}
		} else {
			ChangeSignal(SIGNAL_PAIR_7, MANOEUVRE);

			if (!ZoneOccupee(7))
				ChangeSignal(SIGNAL_IMPAIR_1_0, MANOEUVRE_CLIGNO);
		}

		break;
	case 2:
	case 3:
		if (user_get_aig(0) == AIG_DIRECT) {
			if (CantonOccupe(4)) {
				ChangeSignal(SIGNAL_PAIR_0_1, AVERTISSEMENT);
				ChangeSignal(SIGNAL_IMPAIR_1_0, SEMAPHORE);
			} else {
				ChangeSignal(SIGNAL_PAIR_0_1, VOIE_LIBRE);
				ChangeSignal(SIGNAL_IMPAIR_1_0, VOIE_LIBRE);
			}
		}

		if (CantonOccupe(0)) {
			ChangeSignal(SIGNAL_IMPAIR_5_4, AVERTISSEMENT);
		} else {
			ChangeSignal(SIGNAL_IMPAIR_5_4, VOIE_LIBRE);
			ChangeSignal(SIGNAL_PAIR_4_5, user_get_aig(0) == AIG_DIRECT ? VOIE_LIBRE : CARRE);
		}
		break;
	case 5:
	case 4:
		if (user_get_aig(0) == AIG_DIRECT) {
			if (CantonOccupe(2)) {
				ChangeSignal(SIGNAL_IMPAIR_1_0, AVERTISSEMENT);
				ChangeSignal(SIGNAL_PAIR_0_1, SEMAPHORE);
			} else {
				ChangeSignal(SIGNAL_IMPAIR_1_0, VOIE_LIBRE);
				ChangeSignal(SIGNAL_PAIR_0_1, VOIE_LIBRE);
			}

			if (CantonOccupe(0)) {
				ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);
				ChangeSignal(SIGNAL_IMPAIR_3_2, SEMAPHORE);
			} else {
				ChangeSignal(SIGNAL_PAIR_2_3, VOIE_LIBRE);
				ChangeSignal(SIGNAL_IMPAIR_3_2, VOIE_LIBRE);
			}
		} else
			ChangeSignal(SIGNAL_PAIR_2_3, AVERTISSEMENT);

		break;
	case 7:
		if (user_get_aig(0) == AIG_DEVIE && ! ZoneOccupee(6))
			ChangeSignal(SIGNAL_IMPAIR_1_0, VOIE_LIBRE, RAPPEL_30);
	
		break;
	}
}

// Retourne l'état d'occupation d'un canton. Un canton est formé d'une ou plusieurs zones, en général une zone de ralentissement et une zone d'arrêt.
// Les cantons limitrophes d'une zone d'aiguilles prise en pointe doivent inclure la zone d'aiguille.
// Le rattachement des traversées jonctions et croisements est à effectuer en fonction du tracé.
// Le canton auquel une zone est rattaché pourrait être indiqué dans la table des zones mais le traitement serait plus lent.
//
// Paramètre : numéro de n'importe laquelle des zones du canton. 
// Retour    : true si au moins une des zones du canton est occupé, sinon false.

bool Block::CantonOccupe(byte zone)
{
	switch (zone) {
	case 0:
	case 1:
	case 6:
		// retourner l'état du canton formé des zones 0, 1 et 6.
		return ZoneOccupee(0) || ZoneOccupee(1) || ZoneOccupee(6);
	case 2:
	case 3:
		// retourner l'état du canton formé des zones 2 et 3.
		return ZoneOccupee(2) || ZoneOccupee(3);
	case 4:
	case 5:
		// retourner l'état du canton formé des zones 4 et 5.
		return ZoneOccupee(4) || ZoneOccupee(5);
	default:
		break;
	}

	// retourner l'état des cantons formés d'une seule zone.
	return ZoneOccupee(zone);
}

// ---------- GESTION DES EVENEMENTS PERIODIQUES

// Appelé toutes les 250ms après la régulation de la vitesse des trains.

void Block::OnReguler(void)
{
	 ;
}

#endif
